@@ -1,4 +1,3 @@
-import asyncio
 import itertools
 import logging
 import os
@@ -8,6 +7,7 @@ import urllib.parse
 from typing import Iterable, List, Optional, Tuple, Union
 
 from pygls.lsp.methods import (
+    CODE_ACTION,
     COMPLETION,
     DEFINITION,
     DOCUMENT_LINK,
@@ -18,8 +18,12 @@ from pygls.lsp.methods import (
     TEXT_DOCUMENT_DID_OPEN,
     WORKSPACE_DID_CHANGE_WATCHED_FILES,
 )
-from pygls.server import LanguageServer
 from pygls.lsp.types import (
+    CodeAction,
+    CodeActionKind,
+    CodeActionOptions,
+    CodeActionParams,
+    Command,
     CompletionItem,
     CompletionItemKind,
     CompletionOptions,
@@ -41,9 +45,11 @@ from pygls.lsp.types import (
     TextDocumentPositionParams,
     TextEdit,
 )
+from pygls.lsp.types.window import ShowDocumentParams
+from pygls.server import LanguageServer
 from pygls.workspace import Document
 
-from . import buildout, diagnostic, recipes
+from . import buildout, code_actions, commands, diagnostic, recipes, types, md5sum, utils
 
 server = LanguageServer()
 
@@ -52,8 +58,6 @@ reference_re = re.compile(
     r'\${(?P<section>[-a-zA-Z0-9 ._]*):(?P<option>[-a-zA-Z0-9 ._]+)}')
 logger = logging.getLogger(__name__)
 
-DEBOUNCE_DELAY = 0.3
-
 
 def getOptionValue(
     option: Union[buildout.BuildoutOptionDefinition, str]) -> str:
@@ -65,17 +69,51 @@ def getOptionValue(
   return option.value
 
 
+@utils.singleton_task
 async def parseAndSendDiagnostics(
     ls: LanguageServer,
     uri: str,
 ) -> None:
-  await asyncio.sleep(DEBOUNCE_DELAY)
   diagnostics = []
   async for diag in diagnostic.getDiagnostics(ls, uri):
     diagnostics.append(diag)
   ls.publish_diagnostics(uri, diagnostics)
 
 
+@server.command(commands.COMMAND_OPEN_PYPI_PAGE)
+async def command_open_pypi_page(
+    ls: LanguageServer,
+    args: List[types.OpenPypiPageCommandParams],
+) -> None:
+  await ls.show_document_async(
+      ShowDocumentParams(
+          uri=args[0].url,
+          external=True,
+      ))
+
+
+@server.command(commands.COMMAND_UPDATE_MD5SUM)
+async def command_update_md5sum(
+    ls: LanguageServer,
+    args: List[types.UpdateMD5SumCommandParams],
+) -> None:
+  await md5sum.update_md5sum(ls, args[0])
+
+
+@utils.singleton_task
+@server.feature(
+    CODE_ACTION,
+    CodeActionOptions(resolve_provider=False,
+                      code_action_kinds=[
+                          CodeActionKind.QuickFix,
+                      ]),
+)
+async def lsp_code_action(
+    ls: LanguageServer,
+    params: CodeActionParams) -> Optional[List[Union[Command, CodeAction]]]:
+  return await code_actions.getCodeActions(ls, params)
+
+
 @server.feature(TEXT_DOCUMENT_DID_OPEN)
 async def did_open(
     ls: LanguageServer,
@@ -102,12 +140,12 @@ async def did_change_watched_file(
     buildout.clearCache(change.uri)
 
 
+@utils.singleton_task
 @server.feature(DOCUMENT_SYMBOL)
 async def lsp_symbols(
     ls: LanguageServer,
     params: DocumentSymbolParams,
 ) -> List[DocumentSymbol]:
-  await asyncio.sleep(DEBOUNCE_DELAY)
   symbols: List[DocumentSymbol] = []
 
   parsed = await buildout.parse(
@@ -164,12 +202,12 @@ async def lsp_symbols(
   return symbols
 
 
+@utils.singleton_task
 @server.feature(COMPLETION, CompletionOptions(trigger_characters=["{", ":"]))
 async def lsp_completion(
     ls: LanguageServer,
     params: CompletionParams,
 ) -> Optional[List[CompletionItem]]:
-  await asyncio.sleep(DEBOUNCE_DELAY)
   items: List[CompletionItem] = []
   doc = ls.workspace.get_document(params.text_document.uri)
 
@@ -283,6 +321,8 @@ async def lsp_completion(
   symbol = await parsed.getSymbolAtPosition(params.position)
   logger.debug("getting completions on %s", symbol)
   if symbol:
+    if symbol.kind == buildout.SymbolKind.Comment:
+      return None
     if symbol.kind == buildout.SymbolKind.SectionReference:
       for buildout_section_name, section_items in symbol._buildout.items():
         documentation = '```ini\n{}\n```'.format(
@@ -490,12 +530,12 @@ async def lsp_completion(
   return items
 
 
+@utils.singleton_task
 @server.feature(DEFINITION)
 async def lsp_definition(
     ls: LanguageServer,
     params: TextDocumentPositionParams,
 ) -> List[Location]:
-  await asyncio.sleep(DEBOUNCE_DELAY)
   parsed = await buildout.open(ls, params.text_document.uri)
   if parsed is None:
     return []
@@ -533,6 +573,7 @@ async def lsp_definition(
   return locations
 
 
+@utils.singleton_task
 @server.feature(REFERENCES)
 async def lsp_references(
     server: LanguageServer,
@@ -594,12 +635,12 @@ async def lsp_references(
   return references
 
 
+@utils.singleton_task
 @server.feature(HOVER)
 async def lsp_hover(
     ls: LanguageServer,
     params: TextDocumentPositionParams,
 ) -> Optional[Hover]:
-  await asyncio.sleep(DEBOUNCE_DELAY)
   parsed = await buildout.open(ls, params.text_document.uri)
   if parsed is None:
     return None
@@ -618,12 +659,12 @@ async def lsp_hover(
   return Hover(contents=f'```\n{hover_text}\n```')
 
 
+@utils.singleton_task
 @server.feature(DOCUMENT_LINK)
 async def lsp_document_link(
     ls: LanguageServer,
     params: DocumentLinkParams,
 ) -> List[DocumentLink]:
-  await asyncio.sleep(DEBOUNCE_DELAY)
   links: List[DocumentLink] = []
   uri = params.text_document.uri
   parsed_buildout = await buildout.parse(ls, uri)
